/*
** CSE 322, Lex/Yacc SKELETON: 
**
** jratml.flex: "RATML" formatter, java version, flex half, 
**
** see also jratml.y
**
** W. L. Ruzzo
**
*/

/*
** STUDENTS:
**
**   A: Add your name, uw email and Student ID number below:
**
**      Name:   
**      Email:  @uw.edu
**      Number: 
**
**   B: You're FREE to change anything, but you don't NEED to 
**      change anything except the lex rules section.
*/

/* 
** Notes on the lexer/parser interface:
**
** 1. the TEXT token type passes (through yylval) a java 
**    String containing the text string.
** 2. Token names are defined in jratml.y (* GO READ THIS *), 
**    and written to Parser.java when byacc/j runs.
*/

%%

%byaccj

%{
  /* 
  ** Lexer needs parser (yylval is there), and needs input
  ** stream; constructor below saves both for use in lexer.
  */

  private Parser yyparser;

  public Yylex(java.io.Reader r, Parser yyparser) {
    this(r);
    this.yyparser = yyparser;
  }

/* 
** Table of recognized ratml tags, and associated token codes.  
** Obviously, codes must agree with those in jratml.y.
**
** Students: you don't need to use this, or the associated
** lookup routine, but they might be convenient.
*/

final static String tag[] = {
  "<HTML>",      "<HEAD>",       "<TITLE>",      "<BODY>",     "<UL>",
 "</HTML>",     "</HEAD>",      "</TITLE>",     "</BODY>",    "</UL>",
     "<P>",        "<BR>",          "<LI>"
  };

final static short tagcode[] = {
  Parser.BHTML, Parser.BHEAD, Parser.BTITLE, Parser.BBODY, Parser.BUL, 
  Parser.EHTML, Parser.EHEAD, Parser.ETITLE, Parser.EBODY, Parser.EUL, 
  Parser.P,     Parser.BR,    Parser.LI
  };


/************************************************************* 
** Lookup string in table of recognized tags. If found, return
** associated token code. If not found, it's a TEXT token.
** Binary search would be faster...
*/
short lookup(String str) {
    for(int i=0; i < tag.length; i++) {
        if(str.equals(tag[i])) {
            return tagcode[i];
        }
    }
    return Parser.TEXT;
}

%}

%%
/***  Lex rules below  ***  Students: definitely change them.  ***/
 
[ \t\n]         { ; /* whitespace; don't bother parser */ }

[a-zA-Z]+ |
.               { // text words & punctation to parser 
                  yyparser.yylval = new ParserVal(yytext());
                  // that's "yy*EL*val", not yyval, which will 
                  // give obscure errors...
                  return(Parser.TEXT); // & don't forget token code
                }
 
"</TITLE>"      { // return end title token code, no text.
                  return(Parser.ETITLE); 
                }

