/*
** CSE 322, Lex/Yacc SKELETON: 
**
** jratml.y: "RATML" formatter, java version, yacc half
**
** see also jratml.flex
**
** W. L. Ruzzo
**
*/

/*
** STUDENTS:
**
**   A: Add your name, uw email and student number below:
**
**      Name:   
**      Email:  @uw.edu
**      Number: 
**
**   B: You're FREE to change anything, but you don't NEED to 
**      change anything except the %type declarations and the 
**      yacc rules sections.
*/
%{
  import java.io.*;

  /******************************************************************
  ** The basic data structure for the assignment is the "parse tree"
  ** Node class, plus various subclasses for different markup
  ** elements.  The tree built is not _literally_ a parse tree
  ** according to the grammar, but it reflects the _logically_
  ** important structure of the input, such as nested lists.
  **
  ** The print methods recursively walk the tree and print it in the
  ** obvious left-right order.  Print actions at a given internal node
  ** of the tree can happen before, between, and/or after visiting the
  ** children, depending on the required formatting.
  **
  ** The base class has left & right subtrees, and can be used quite
  ** flexibly, but in most cases it is used to build lists of subclass
  ** elements.  E.g., the word list in <TITLE> might use the left
  ** pointers to build a null-terminated linked list of Nodes, whose
  ** right pointers each connect to a NodeText object containing one
  ** word.
  **
  ** Note that the left-right print order means that these trees:
  **
  **           / \        /\
  **          /   \      /  \
  **        /  \   C    C  / \
  **      / \   B         B  / \
  **   null  A              A  null
  **
  ** (where A,B,C represent NodeText objects and the other nodes are
  ** Node objects) will print "A B C" and "C B A", resp.  I.e., to
  ** have the deepest leaf in a list print first, use left pointers to
  ** build the "list" (the unlabeled nodes above); use right pointers
  ** if the deepest leaf should print last.
  **
  */
  class Node {
    Node left; 
    Node right;
    Node(Object ileft, Object iright){
      left  = (Node) ileft;  //casts here save many others in rule actions
      right = (Node) iright;
    }
    
    void print(int margin){
      if(left  != null) left.print(margin);
      if(right != null) right.print(margin);
    }
    
    // data and helper functions for rendering formatted output:

    static String line = "";   // output line buffer
    static int tempIndent = 0; // temporary indent, used for <LI>

    /******************************************************************
    ** blank: print a blank line, forcing out any partially full line
    ** if necessary
    */
    void blank(int margin) {
      cbr(margin);
      System.out.println();
    }

    /******************************************************************
    ** conditional line break: force out any partially full line if
    ** necessary; nothing to do if line buffer empty
    */
    void cbr(int margin){
      if(line.length() != 0) {
	br(margin);
      }
    }

    /******************************************************************
    ** line break: output current line, indented as needed, and reset
    ** for next line.
    */
    void br(int margin){
      int slack =  margin + tempIndent;
      while(slack-- > 0) {
	System.out.print(" ");
      }
      System.out.println(line);
      line = "";
      tempIndent = 0;
    }
  
    /******************************************************************
    ** put_text: add next word to text out stream; do line break first
    ** if it would otherwise extend past right margin.
    **
    ** Note: if next word plus margin is longer than 80 characters, it
    ** will be printed by itself on a long line.  Lines otherwise
    ** should be <= 80 characters.
    */
    void put_text(int margin, String text){
      // need between-word space?
      int sp = (line.length() > 0) ? 1 : 0 ; 
      if(margin+tempIndent+sp+line.length()+text.length() > 80){
	br(margin);
      } 
      if(line.length() > 0) {
	line += " ";
      }
      line += text;
    }
  }

  // Node subclasses:

  // top level of parse tree; left is title, right is body
  class NodePage extends Node {
    NodePage(Object left, Object right){super(left,right);}
    void print(){
      if(left  != null) left.print(0);
      blank(0); // blank line after title
      if(right != null) right.print(0);
      cbr(0);   // force out final (partial) line
    }
  }

  // leaf node holding a TEXT token
  class NodeTEXT extends Node {
    String text;
    NodeTEXT(Object istring){
      super(null, null);
      text = (String) istring;
    }
    void print(int margin){
      put_text(margin, text);
    }
  }

  // leaf node for <BR> tag
  class NodeBR extends Node {
    NodeBR(){super(null,null);}
    void print(int margin){
      cbr(margin);
    }
  }

  // leaf node for <P> tag
  class NodeP extends Node {
    NodeP(){super(null,null);}
    void print(int margin){
      blank(margin);
    }
  }

  // node for <UL></UL> wrapper
  class NodeUL extends Node {
    NodeUL(Object left){super(left, null);}
    void print(int margin){
      cbr(margin);
      if(left != null){
	left.print(margin+4);
	cbr(margin+4);
      }
    }
  }

  // node for <LI> tag
  class NodeLI extends Node {
    NodeLI(Object left){super(left,null);}
    void print(int margin){
      cbr(margin);
      tempIndent = -2;  // out-dent the star marking a list item
      put_text(margin, "*");
      if(left != null) left.print(margin);
    }
  }

%}

/* 
** Declare all lex tokens; TEXT passes a string value (hence the
** <sval>); token code alone suffices for the others.  Token names
** correspond to html tags:
**   "B" tokens are begining tags of wrappers,
**   "E" tokens are corresponding end tags,
**   "P", "BR" etc. are assumed NOT to have matching end tag.
**
** STUDENTS: my solution uses these, but you are free to change them.
*/
%token BHTML BHEAD BTITLE BBODY BUL
%token EHTML EHEAD ETITLE EBODY EUL
%token P BR LI 
%token <sval> TEXT

/*
** Declare all grammar nonterminals; all carry an Object
** type (hence the <obj>) via their $$, etc. variables. 
** First one/first rule define the start symbol.
*/
%type <obj> page word words w4 aa bb

%%

/*
** Yacc rules section.
**
** Students: you definitely need to change this (but read it first).
**
** This sample grammer says (roughly) that a page has one or more
** ETITLE tokens, plus optional TEXT tokens.  It serves no purpose
** other than to show an example of grammar/lex/parse tree/formatter
** interfaces.
**
** However, before you delete it, you should run it, and look at it
** (and the code above) to figure out, in the output generated from
** test.ratml, (a) why the first line comes out backwards, (b) why
** "</HEAD>" didn't get printed, (c) how you could fix those, and (d)
** why one blank line was printed.
*/

page:   aa w4 bb                { $$ = new NodePage($1, $3);
                                  ((NodePage)$$).print();
                                }
      ;
w4:     word word word word     { $$ = null; }
      ;

aa:     words ETITLE            { $$ = new Node($1, null); } 
      ;

words:  words word              { $$ = new Node($2, $1); }
      | /* empty */             { $$ = null; /* don't forget null; obscure bugs otherwise */ }
      ; 

word:   TEXT                    { $$ = new NodeTEXT($1); }
      ;

bb:     bb word                 { $$ = new Node($1, $2); }
      | bb ETITLE               { $$ = $1; } 
      | /* empty */             { $$ = null; }
      ;

%%

  /*
  ** Initialization and main program.
  **
  ** Nothing here needs to change, but you're free to do so.  The 0/1
  ** command line agrument may help with your lex debugging.
  */

  private Yylex lexer;

  private int yylex () {
    int yyl_return = -1;
    try {
      yylval = new ParserVal(0);
      yyl_return = lexer.yylex();
    }
    catch (IOException e) {
      System.err.println("IO error :"+e);
    }
    return yyl_return;
  }

  public void yyerror (String error) {
    System.err.println ("Error: " + error);
  }

  public Parser(Reader r) {
    lexer = new Yylex(r, this);
  }

  public static void main(String args[]) throws IOException {
    System.out.println("BYACC/Java + JFlex Ratml demo");
    System.out.println("=============================");
    if ( args.length < 1 ) {
      System.out.println("Usage: java Parser 0/1 filename");
      System.out.println("where 0 means run yacc; 1 means just dump lex tokens");
    }
    Parser yyparser = new Parser(new FileReader(args[1]));
    yyparser.yydebug = false ; //true turns on yacc debug printouts;
    if(args[0].equals("1")){
      System.out.println("token dump:");
      //dump tokens
      int tok;
      while((tok=yyparser.lexer.yylex()) > 0){
	System.out.print("Tok="+tok);
	if(null != yyparser.yylval && null != yyparser.yylval.sval){
	  System.out.print(","+yyparser.yylval.sval);
	}
	System.out.println('.');
      }
    }
    else {
      yyparser.yyparse();
    }
  }
